# Python Version: 3.x
# -*- coding: utf-8 -*-
import collections
import io
import re
import zipfile
from typing import *

import onlinejudge._implementation.format_utils
import onlinejudge._implementation.logging as log
from onlinejudge.type import *


class SampleZipper(object):
    def __init__(self):
        self._testcases = []  # List[TestCase]
        self._dangling = None  # Optional[Tuple(str, bytes)]

    def add(self, content: bytes, name: str) -> None:
        if self._dangling is None:
            if re.search('output', name, re.IGNORECASE) or re.search('出力', name):
                log.warning('strange name for input string: %s', name)
            self._dangling = (name, content)
        else:
            if re.search('input', name, re.IGNORECASE) or re.search('入力', name):
                if not (re.search('output', name, re.IGNORECASE) or re.search('出力', name)):  # to ignore titles like "Output for Sample Input 1"
                    log.warning('strange name for output string: %s', name)
            index = len(self._testcases)
            input_name, input_content = self._dangling
            self._testcases += [TestCase('sample-{}'.format(index + 1), input_name, input_content, name, content)]
            self._dangling = None

    def get(self) -> List[TestCase]:
        if self._dangling is not None:
            log.error('dangling sample string: %s', self._dangling[1])
            raise SampleParseError
        return self._testcases


def extract_from_files(files: Iterator[Tuple[str, bytes]], format: str = '%s.%e', out: str = 'out') -> List[TestCase]:
    """
    :param out: is the extension for output files. This is used when the zip-file contains files like `sample-1.ans` instead of `sample-1.out`.
    """

    table = {
        's': r'[^/]+',
        'e': r'(in|{})'.format(out),
    }
    names = collections.defaultdict(dict)  # type: Dict[str, Dict[str, Tuple[str, bytes]]]
    for filename, content in files:
        m = onlinejudge._implementation.format_utils.percentparse(filename, format, table)
        assert m
        assert m['e'] not in names[m['s']]
        names[m['s']][m['e']] = (filename, content)
    testcases = []  # type: List[TestCase]
    for name in sorted(names.keys()):
        data = names[name]
        if 'in' not in data or out not in data:
            log.error('dangling sample found: %s', str(data))
            assert False
        else:
            testcases += [TestCase(name, *data['in'], *data[out])]
    return testcases


def extract_from_zip(zip_data: bytes, format: str, out: str = 'out') -> List[TestCase]:
    def iterate():
        with zipfile.ZipFile(io.BytesIO(zip_data)) as fh:
            for filename in fh.namelist():
                if filename.endswith('/'):  # TODO: use `fh.getinfo(filename).is_dir()` after we stop supporting Python 3.5
                    continue
                yield filename, fh.read(filename)

    return extract_from_files(iterate(), format=format, out=out)



# Python Version: 3.x
import json
import os
import pathlib
from typing import *

import requests.exceptions

import onlinejudge
import onlinejudge._implementation.download_history
import onlinejudge._implementation.format_utils as format_utils
import onlinejudge._implementation.logging as log
import onlinejudge._implementation.utils as utils
import onlinejudge.type

if TYPE_CHECKING:
    import argparse


def convert_sample_to_dict(sample: onlinejudge.type.TestCase) -> Dict[str, str]:
    data = {}  # type: Dict[str, str]
    data["name"] = sample.name
    data["input"] = sample.input_data.decode()
    if sample.output_data is not None:
        data["output"] = sample.output_data.decode()
    return data


def download(args: 'argparse.Namespace') -> None:
    # prepare values
    problem = onlinejudge.dispatch.problem_from_url(args.url)
    if problem is None:
        raise requests.exceptions.InvalidURL('The contest "%s" is not supported' % args.url)
    is_default_format = args.format is None and args.directory is None  # must be here since args.directory and args.format are overwritten
    if args.directory is None:
        args.directory = pathlib.Path('test')
    if args.format is None:
        args.format = '%b.%e'

    # get samples from the server
    with utils.with_cookiejar(utils.new_session_with_our_user_agent(), path=args.cookie) as sess:
        if args.system:
            samples = problem.download_system_cases(session=sess)
        else:
            samples = problem.download_sample_cases(session=sess)

    # append the history for submit command
    if not args.dry_run and is_default_format:
        history = onlinejudge._implementation.download_history.DownloadHistory()
        history.add(problem)

    # write samples to files
    for i, sample in enumerate(samples):
        log.emit('')
        log.info('sample %d', i)
        for ext in ['in', 'out']:
            data = getattr(sample, ext + 'put_data')
            if data is None:
                continue
            name = sample.name
            table = {}
            table['i'] = str(i + 1)
            table['e'] = ext
            table['n'] = name
            table['b'] = os.path.basename(name)
            table['d'] = os.path.dirname(name)
            path = args.directory / format_utils.percentformat(args.format, table)  # type: pathlib.Path
            log.status('%sput: %s', ext, name)
            if not args.silent:
                log.emit(utils.snip_large_file_content(data, limit=40, head=20, tail=10, bold=True))
            if args.dry_run:
                continue
            if path.exists():
                raise FileExistsError('Failed to download since file already exists: ' + str(path))
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open('wb') as fh:
                fh.write(data)
            log.success('saved to: %s', path)

    # print json
    if args.json:
        print(json.dumps(list(map(convert_sample_to_dict, samples))))



# Python Version: 3.x
import contextlib
import os
import subprocess
import sys
from typing import *

import onlinejudge._implementation.logging as log

if TYPE_CHECKING:
    import argparse


@contextlib.contextmanager
def fifo() -> Generator[Tuple[Any, Any], None, None]:
    fdr, fdw = os.pipe()
    fhr = os.fdopen(fdr, 'r')
    fhw = os.fdopen(fdw, 'w')
    yield fhr, fhw
    fhw.close()
    fhr.close()
    # os.close(fdw), os.close(fdr) are unnecessary


def test_reactive(args: 'argparse.Namespace') -> None:
    with fifo() as (fhr1, fhw1):
        with fifo() as (fhr2, fhw2):
            with subprocess.Popen(args.command, shell=True, stdin=fhr2, stdout=fhw1, stderr=sys.stderr) as proc1:
                with subprocess.Popen(args.judge, shell=True, stdin=fhr1, stdout=fhw2, stderr=sys.stderr) as proc2:
                    proc1.communicate()
                    proc2.communicate()
                    if proc1.returncode != 0:
                        log.failure('RE: solution returns %d', proc1.returncode)
                    if proc2.returncode == 0:
                        log.success('AC')
                    else:
                        log.failure('WA: judge returns %d', proc2.returncode)



# Python Version: 3.x
"""
the module containing base types

:note: Some methods are not implemented in subclasses.
    Please check the definitions of subclasses under :py:mod:`onlinejudge.service`.
"""

import datetime
from abc import ABC, abstractmethod
from typing import Callable, Iterator, List, NamedTuple, NewType, Optional, Sequence, Tuple

import requests

CredentialsProvider = Callable[[], Tuple[str, str]]


class LoginError(RuntimeError):
    pass


class Service(ABC):
    def login(self, *, get_credentials: CredentialsProvider, session: Optional[requests.Session] = None) -> None:
        """
        :param get_credentials: returns a tuple of (username, password)
        :raises LoginError:
        """
        raise NotImplementedError

    def get_url_of_login_page(self) -> str:
        """
        .. versionadded:: 7.0.0
        """
        raise NotImplementedError

    def is_logged_in(self, *, session: Optional[requests.Session] = None) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_url(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_name(self) -> str:
        """
        example:

        -   `AtCoder`
        -   `Codeforces`
        -   `PKU JudgeOnline`

        :note: If you want something like identifier (e.g. `atcoder`, `codeforces` or `poj`), you can use a domain obtained from :py:meth:`get_url`.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return '{}.from_url({})'.format(self.__class__.__name__, repr(self.get_url()))

    def __eq__(self, other) -> bool:
        return self.__class__ == other.__class__ and self.get_url() == other.get_url()

    @classmethod
    @abstractmethod
    def from_url(self, s: str) -> Optional['Service']:
        pass

    def iterate_contests(self, *, session: Optional[requests.Session] = None) -> Iterator['Contest']:
        """
        .. versionadded:: 7.0.0
        """
        raise NotImplementedError


TestCase = NamedTuple('TestCase', [
    ('name', str),
    ('input_name', str),
    ('input_data', bytes),
    ('output_name', str),
    ('output_data', bytes),
])
"""
.. versionchanged:: 7.0.0
"""

LanguageId = NewType('LanguageId', str)
"""
:note: This is just a :py:class:`NewType` -ed :py:class:`str` not, but you should not use this other than a label.
"""

Language = NamedTuple('Language', [
    ('id', LanguageId),
    ('name', str),
])
"""
:ivar id: :py:class:`LanguageId`
:ivar name: :py:class:`str`
"""


class NotLoggedInError(RuntimeError):
    pass


class SampleParseError(RuntimeError):
    """
    .. versionadded:: 7.0.0
    """
    pass


class SubmissionError(RuntimeError):
    pass


class DownloadedData(ABC):
    """
    :note: :py:class:`DownloadedData` and its subclasses represent contents which are obtained by network access. The values may depends your session.
           :py:class:`DownloadedData` とそのサブクラスは、ネットワークアクセスの結果得られるようなデータを表現します。その値はログイン状況などにより接続のたびに変化することがあります。

    .. versionadded:: 7.0.0
    """
    @property
    @abstractmethod
    def url(self) -> str:
        raise NotImplementedError

    @property
    def json(self) -> Optional[bytes]:
        return None

    @property
    def html(self) -> Optional[bytes]:
        return None

    @property
    def timestamp(self) -> Optional[datetime.datetime]:
        return None

    @property
    def session(self) -> Optional[requests.Session]:
        return None

    @property
    def response(self) -> Optional[requests.Response]:
        return None


class ContestData(DownloadedData):
    """
    .. versionadded:: 7.0.0
    """
    def url(self) -> str:
        return self.contest.get_url()

    @property
    @abstractmethod
    def contest(self) -> 'Contest':
        raise NotImplementedError

    @property
    def service(self) -> Service:
        return self.contest.get_service()

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError


class Contest(ABC):
    """
    :note: :py:class:`Contest` represents just a URL of a contest, without the data of the contest.

    .. versionadded:: 7.0.0
    """
    def list_problems(self, *, session: Optional[requests.Session] = None) -> Sequence['Problem']:
        raise NotImplementedError

    def download_data(self, *, session: Optional[requests.Session] = None) -> ContestData:
        raise NotImplementedError

    def iterate_submissions(self, *, session: Optional[requests.Session] = None) -> Iterator['Submission']:
        """
        .. versionadded:: 7.0.0
        """
        raise NotImplementedError

    @abstractmethod
    def get_url(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_service(self) -> Service:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_url(self, s: str) -> Optional['Contest']:
        pass


class ProblemData(DownloadedData):
    """
    .. versionadded:: 7.0.0
    """
    def url(self) -> str:
        return self.problem.get_url()

    @property
    @abstractmethod
    def problem(self) -> 'Problem':
        raise NotImplementedError

    @property
    def contest(self) -> Contest:
        return self.problem.get_contest()

    @property
    def service(self) -> Service:
        return self.problem.get_service()

    @property
    @abstractmethod
    def name(self) -> str:
        """
        for example of :py:class:`Problem`:

        -   `器物損壊！高橋君`
        -   `AtCoDeerくんと変なじゃんけん / AtCoDeer and Rock-Paper`
        -   `Xor Sum`
        """
        raise NotImplementedError

    @property
    def sample_cases(self) -> Optional[List[TestCase]]:
        raise NotImplementedError


class Problem(ABC):
    """
    :note: :py:class:`Problem` represents just a URL of a problem, without the data of the problem.
           :py:class:`Problem` はちょうど問題の URL のみを表現します。キャッシュや内部状態は持ちません。
    """
    @abstractmethod
    def download_sample_cases(self, *, session: Optional[requests.Session] = None) -> List[TestCase]:
        """
        :raises SampleParseError:

        .. versionchanged:: 7.0.0
        """
        raise NotImplementedError

    def download_system_cases(self, *, session: Optional[requests.Session] = None) -> List[TestCase]:
        """
        :raises NotLoggedInError:
        """
        raise NotImplementedError

    def submit_code(self, code: bytes, language_id: LanguageId, *, filename: Optional[str] = None, session: Optional[requests.Session] = None) -> 'Submission':
        """
        :param code:
        :arg language_id: :py:class:`LanguageId`
        :raises NotLoggedInError:
        :raises SubmissionError:
        """
        raise NotImplementedError

    def get_available_languages(self, *, session: Optional[requests.Session] = None) -> List[Language]:
        raise NotImplementedError

    @abstractmethod
    def get_url(self) -> str:
        raise NotImplementedError

    def get_contest(self) -> Contest:
        raise NotImplementedError

    @abstractmethod
    def get_service(self) -> Service:
        raise NotImplementedError

    def download_data(self, *, session: Optional[requests.Session] = None) -> ProblemData:
        """
        .. versionadded:: 7.0.0
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return '{}.from_url({})'.format(self.__class__.__name__, repr(self.get_url()))

    def __eq__(self, other) -> bool:
        return self.__class__ == other.__class__ and self.get_url() == other.get_url()

    @classmethod
    @abstractmethod
    def from_url(self, s: str) -> Optional['Problem']:
        pass


class SubmissionData(DownloadedData):
    """
    .. versionadded:: 7.0.0
    """
    def url(self) -> str:
        return self.submission.get_url()

    @property
    @abstractmethod
    def submission(self) -> 'Submission':
        raise NotImplementedError

    @property
    def service(self) -> Service:
        return self.submission.get_service()

    @property
    def source_code(self) -> bytes:
        raise NotImplementedError

    @property
    @abstractmethod
    def status(self) -> str:
        raise NotImplementedError


class Submission(ABC):
    def download_data(self, *, session: Optional[requests.Session] = None) -> SubmissionData:
        """
        .. versionadded:: 7.0.0
        """
        raise NotImplementedError

    @abstractmethod
    def get_url(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def download_problem(self, *, session: Optional[requests.Session] = None) -> Problem:
        raise NotImplementedError

    def download_contest(self) -> Contest:
        return self.download_problem().get_contest()

    @abstractmethod
    def get_service(self) -> Service:
        raise NotImplementedError

    def __repr__(self) -> str:
        return '{}.from_url({})'.format(self.__class__.__name__, repr(self.get_url()))

    def __eq__(self, other) -> bool:
        return self.__class__ == other.__class__ and self.get_url() == other.get_url()

    @classmethod
    @abstractmethod
    def from_url(cls, s: str) -> Optional['Submission']:
        raise NotImplementedError
